# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2023 Ronny Scharf-W. <ronny.scharf08@gmail.com>         *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, cast
import importlib
import warnings
import inspect
import time

import awkward as ak
import numpy as np

# noinspection PyUnresolvedReferences
import FreeCAD
import Part
import Points  # noqa

import PySide2.QtCore as QtCore
import PySide2.QtWidgets as QtWidgets

from utils import simplified_array_structure, simplify_array
from nested_data import NestedData
from utils import record_structure, flatten_record
from node_item import NodeItem
from input_widgets import OptionBoxWidget
from sockets.shape_none import ShapeNone

if TYPE_CHECKING:
    from socket_widget import SocketWidget


DEBUG = True


class Boolean(NodeItem):
    REG_NAME: str = "Boolean"

    def __init__(self, pos: tuple, undo_stack: QtWidgets.QUndoStack, name: str = REG_NAME,
                 parent: Optional[QtWidgets.QGraphicsItem] = None) -> None:
        super().__init__(pos, undo_stack, name, parent)

        # Option combo box
        self._option_box: OptionBoxWidget = OptionBoxWidget(undo_stack)
        self._option_box.setFocusPolicy(QtCore.Qt.NoFocus)
        self._option_box.setMinimumWidth(5)
        self._option_box.addItems(["Union", "Subtraction", "Intersection", "Section", "Mass Union"])
        for option_idx in range(self._option_box.count()):
            self._option_box.model().setData(self._option_box.model().index(option_idx, 0), QtCore.QSize(160, 24),
                                             QtCore.Qt.SizeHintRole)

        item_list_view: QtWidgets.QListView = cast(QtWidgets.QListView, self._option_box.view())
        item_list_view.setSpacing(2)
        self._content_widget.hide()
        self._content_layout.addWidget(self._option_box)
        self._content_widget.show()

        # Socket widgets
        self._socket_widgets: list[SocketWidget] = [
            ShapeNone(undo_stack=self._undo_stack, name="Shape A", content_value="<No Input>", is_input=True,
                      parent_node=self),
            ShapeNone(undo_stack=self._undo_stack, name="Shape B", content_value="<No Input>", is_input=True,
                      parent_node=self),
            ShapeNone(undo_stack=self._undo_stack, name="Out", content_value="<No Input>", is_input=False,
                      parent_node=self)
        ]

        # Listeners
        cast(QtCore.SignalInstance, self._option_box.currentIndexChanged).connect(self.update_socket_widgets)

    def update_socket_widgets(self) -> None:
        # Hack to prevent cyclic imports
        add_socket_cmd_cls: type = getattr(importlib.import_module("undo_commands"), "AddSocketCommand")
        remove_socket_cmd_cls: type = getattr(importlib.import_module("undo_commands"), "RemoveSocketCommand")
        remove_edge_cmd_cls: type = getattr(importlib.import_module("undo_commands"), "RemoveEdgeCommand")
        set_op_idx_cmd_cls: type = getattr(importlib.import_module("undo_commands"), "SetOptionIndexCommand")
        emit_dag_changed_cmd_cls: type = getattr(importlib.import_module("undo_commands"), "EmitDagChangedCommand")

        last_option_index: int = self._option_box.last_index
        current_option_name: str = self._option_box.currentText()
        current_option_index: int = self._option_box.currentIndex()
        input_widget_count: int = len(self.input_socket_widgets)

        self._undo_stack.beginMacro("Changes option box")
        self._undo_stack.push(emit_dag_changed_cmd_cls(self.scene(), self))

        if current_option_name in ("Union", "Subtraction", "Intersection", "Section"):
            while input_widget_count < 2:
                new_socket_widget: ShapeNone = ShapeNone(undo_stack=self._undo_stack, name="Shape B",
                                                         content_value="<No Input>", is_input=True, parent_node=self)
                insert_idx: int = len(self.input_socket_widgets)
                self._undo_stack.push(add_socket_cmd_cls(self, new_socket_widget, insert_idx))
                input_widget_count += 1

        else:
            while input_widget_count > 1:
                remove_idx: int = len(self.input_socket_widgets) - 1
                remove_socket: SocketWidget = self._socket_widgets[remove_idx]
                for edge in remove_socket.pin.edges:
                    self._undo_stack.push(remove_edge_cmd_cls(self.scene(), edge, True))
                self._undo_stack.push(remove_socket_cmd_cls(self, remove_idx))
                input_widget_count -= 1

        self._undo_stack.push(
            set_op_idx_cmd_cls(self, self._option_box, last_option_index, current_option_index)
        )
        self._undo_stack.push(emit_dag_changed_cmd_cls(self.scene(), self, on_redo=True))
        self._undo_stack.endMacro()

    # --------------- Node eval methods ---------------

    def eval_0(self, *args) -> ak.Array:
        cache_idx: int = int(inspect.stack()[0][3].split("_")[-1])

        if self._is_invalid or self._cache[cache_idx] is None:
            with warnings.catch_warnings():
                warnings.filterwarnings("error")
                try:
                    try:
                        shape_a: NestedData = self.input_data(0, args)

                        if DEBUG:
                            a: float = time.time()

                        if len(args) == 1:
                            simple_shapes, struct_shapes = (simplify_array(shape_a.structure),
                                                            simplified_array_structure(shape_a.structure))
                            flat_data: list[Part.Shape] = []
                            if type(struct_shapes) is int:
                                first: Part.Shape = shape_a.data[0]
                                if len(shape_a.data) > 1:
                                    rest: list[Part.Shape] = shape_a.data[1:]
                                    flat_data.append(first.multiFuse(rest))
                                else:
                                    flat_data.append(first)
                            else:
                                shapes: np.ndarray = np.array(shape_a.data, dtype="object")
                                for simple_idx in simple_shapes:
                                    sub_shapes: np.ndarray = shapes[simple_idx]
                                    first: Part.Shape = sub_shapes[0]
                                    if len(sub_shapes) > 1:
                                        rest: list[Part.Shape] = sub_shapes[1:]
                                        flat_data.append(first.multiFuse(rest))
                                    else:
                                        flat_data.append(first)

                            result: NestedData = NestedData(
                                data=flat_data,
                                structure=struct_shapes if type(struct_shapes) == ak.Array else ak.Array([0])
                            )

                        if len(args) == 2:
                            shape_b: NestedData = self.input_data(1, args)

                            broadcasted_params: ak.Array = ak.zip(
                                {"shape_a": shape_a.structure, "shape_b": shape_b.structure}
                            )
                            flat_params: ak.Array = flatten_record(nested_record=broadcasted_params, as_tuple=True)

                            flat_data: list[Part.Shape] = []
                            for param_tuple in flat_params:
                                copy_a: Part.Shape = Part.Shape(shape_a.data[param_tuple["0"]])
                                copy_b: Part.Shape = Part.Shape(shape_b.data[param_tuple["1"]])

                                if self._option_box.currentText() == "Union":
                                    flat_data.append(copy_a.fuse(copy_b))

                                elif self._option_box.currentText() == "Subtraction":
                                    flat_data.append(copy_a.cut(copy_b))

                                elif self._option_box.currentText() == "Intersection":
                                    flat_data.append(copy_a.common(copy_b))

                                elif self._option_box.currentText() == "Section":
                                    flat_data.append(copy_a.section(copy_b))

                            result: NestedData = NestedData(
                                data=flat_data,
                                structure=record_structure(broadcasted_params)
                            )

                        self._is_dirty: bool = False
                        self._is_invalid: bool = False
                        self._cache[cache_idx] = self.output_data(0, result)

                        if DEBUG:
                            b: float = time.time()
                            print("Boolean executed in", "{number:.{digits}f}".format(number=1000 * (b - a), digits=2),
                                  "ms")

                    except Exception as e:
                        self._is_dirty: bool = True
                        print(e)
                except Warning as e:
                    self._is_dirty: bool = True
                    print(e)

        return self._cache[cache_idx]

# --------------- Serialization ---------------

    def __getstate__(self) -> dict:
        data_dict: dict = super().__getstate__()
        data_dict["Option Idx"] = self._option_box.currentIndex()
        return data_dict

    def __setstate__(self, state: dict):
        super().__setstate__(state)
        self._option_box.blockSignals(True)
        self._option_box.setCurrentIndex(state["Option Idx"])
        self._option_box.blockSignals(False)
        self.update()
